@@ -4,6 +4,17 @@
 
 -- | A thread manager including a time manager.
 --   The manager has responsibility to kill managed threads.
+--
+-- Because this is based on the accompanying "System.TimeManager" module,
+-- the same caveats apply:
+--
+--   * Only works for GHC.
+--   * Only works with a threaded runtime.
+--   * Users of older versions should check the current semantics.
+--   * Using 32-bit systems means the max timeout is @'maxBound' :: Int@
+--     (2147483647) microseconds, which is less than 36 minutes.
+--   * Using the same 'Handle' in different threads might cause issues in some
+--     edge cases. (i.e. using cancel/pause in one thread, and resume in another)
 module System.ThreadManager (
     ThreadManager,
     newThreadManager,
@@ -36,7 +47,7 @@ import Control.Exception (Exception (..), SomeException (..))
 import qualified Control.Exception as E
 import Control.Monad (unless, void)
 import Data.Foldable (forM_)
-import Data.IORef
+import Data.IORef (IORef, atomicModifyIORef', newIORef)
 import Data.Map.Strict (Map)
 import qualified Data.Map.Strict as Map
 import Data.Word (Word64)
@@ -65,10 +76,12 @@ data ManagedThread = ManagedThread (Weak ThreadId) (IORef Bool)
 
 ----------------------------------------------------------------
 
--- | Starting a thread manager.
---   Its action is initially set to 'return ()' and should be set
---   by 'setAction'. This allows that the action can include
---   the manager itself.
+-- | Create a thread manager.
+--
+-- To create a 'ThreadManager', you will first have to create a
+-- 'T.Manager' from the "System.TimeManager" module.
+--
+-- You can use either 'System.TimeManager.initialize' or 'System.TimeManager.withManager'.
 newThreadManager :: T.Manager -> IO ThreadManager
 newThreadManager timmgr = ThreadManager timmgr <$> newTVarIO Map.empty
 
@@ -84,6 +97,10 @@ instance Exception KilledByThreadManager where
 
 -- | Stopping the manager.
 --
+-- @
+-- stopAfter threadManager action cleanup
+-- @
+--
 -- The action is run in the scope of an exception handler that catches all
 -- exceptions (including asynchronous ones); this allows the cleanup handler
 -- to cleanup in all circumstances. If an exception is caught, it is rethrown
@@ -123,14 +140,25 @@ forkManaged mgr label io =
 
 -- | Like 'forkManaged', but run action with exceptions masked
 forkManagedUnmask
-    :: ThreadManager -> String -> ((forall x. IO x -> IO x) -> IO ()) -> IO ()
+    :: ThreadManager
+    -> String
+    -- ^ Thread name
+    -> ((forall x. IO x -> IO x) -> IO ())
+    -- ^ Action with unmask argument
+    -> IO ()
 forkManagedUnmask (ThreadManager _timmgr var) label io =
     void $ E.mask_ $ forkIOWithUnmask $ \unmask -> E.handle ignore $ do
         labelMe label
         E.bracket (setup var) (clear var) $ \_ -> io unmask
 
 -- | Fork a managed thread with a handle created by a timeout manager.
-forkManagedTimeout :: ThreadManager -> String -> (T.Handle -> IO ()) -> IO ()
+forkManagedTimeout
+    :: ThreadManager
+    -> String
+    -- ^ Thread name
+    -> (T.Handle -> IO ())
+    -- ^ Action with timeout handle
+    -> IO ()
 forkManagedTimeout (ThreadManager timmgr var) label io =
     void $ forkIO $ do
         labelMe label
@@ -140,26 +168,37 @@ forkManagedTimeout (ThreadManager timmgr var) label io =
     ex = KilledByThreadManager Nothing
 
 -- | Fork a managed thread with a cleanup function.
-forkManagedFinally :: ThreadManager -> String -> IO () -> IO () -> IO ()
-forkManagedFinally mgr label io final = E.mask $ \restore ->
-    forkManaged
-        mgr
-        label
-        (E.try (restore io) >>= \(_ :: Either E.SomeException ()) -> final)
+forkManagedFinally
+    :: ThreadManager
+    -> String
+    -- ^ Thread name
+    -> IO ()
+    -- ^ Action
+    -> IO ()
+    -- ^ Cleanup function
+    -> IO ()
+forkManagedFinally mgr label io final =
+    forkManagedUnmask mgr label $ \restore ->
+        E.try (restore io) >>= \(_ :: Either E.SomeException ()) -> final
 
 -- | Fork a managed thread with a handle created by a timeout manager
 -- and with a cleanup function.
 forkManagedTimeoutFinally
-    :: ThreadManager -> String -> (T.Handle -> IO ()) -> IO () -> IO ()
+    :: ThreadManager
+    -> String
+    -- ^ Thread name
+    -> (T.Handle -> IO ())
+    -- ^ Action with timeout handle
+    -> IO ()
+    -- ^ Cleanup function
+    -> IO ()
 forkManagedTimeoutFinally mgr label io final = E.mask $ \restore ->
-    forkManagedTimeout
-        mgr
-        label
-        (\th -> E.try (restore $ io th) >>= \(_ :: Either E.SomeException ()) -> final)
+    forkManagedTimeout mgr label $ \th ->
+        E.try (restore $ io th) >>= \(_ :: Either E.SomeException ()) -> final
 
 setup :: TVar (Map Key ManagedThread) -> IO (Key, Weak ThreadId, IORef Bool)
 setup var = do
-    (wtid, n) <- myWeakThradId
+    (wtid, n) <- myWeakThreadId
     ref <- newIORef False
     let ent = ManagedThread wtid ref
     -- asking to throw KilledByThreadManager to me
@@ -184,21 +223,21 @@ clear var (n, _, _) = atomically $ modifyTVar' var $ Map.delete n
 ignore :: KilledByThreadManager -> IO ()
 ignore (KilledByThreadManager _) = return ()
 
--- | Wait until all managed thread are finished.
+-- | Wait until all managed threads are finished.
 waitUntilAllGone :: ThreadManager -> IO ()
-waitUntilAllGone (ThreadManager _timmgr var) = atomically $ do
-    m <- readTVar var
-    check (Map.size m == 0)
+waitUntilAllGone tm =
+    atomically $
+        isAllGone tm >>= check
 
+-- | STM action that checks if all managed threads are finished.
 isAllGone :: ThreadManager -> STM Bool
-isAllGone (ThreadManager _timmgr var) = do
-    m <- readTVar var
-    return (Map.size m == 0)
+isAllGone (ThreadManager _timmgr var) =
+    Map.null <$> readTVar var
 
 ----------------------------------------------------------------
 
-myWeakThradId :: IO (Weak ThreadId, Key)
-myWeakThradId = do
+myWeakThreadId :: IO (Weak ThreadId, Key)
+myWeakThreadId = do
     tid <- myThreadId
     wtid <- mkWeakThreadId tid
     let n = fromThreadId tid
@@ -209,6 +248,8 @@ labelMe l = do
     tid <- myThreadId
     labelThread tid l
 
+-- | Registering a 'T.TimeoutAction' and unregister its 'T.Handle'
+--   when the body action is finished.
 withHandle
     :: ThreadManager -> T.TimeoutAction -> (T.Handle -> IO a) -> IO a
 withHandle (ThreadManager timmgr _) = T.withHandle timmgr
